import sys

sys.path.append('.')
from easydict import EasyDict
import yaml
import numpy as np
from scipy.spatial.transform import Rotation


def format_cal_data(cal_data):
    """
    generate extrinsic mat, camera intrinsic mat and distort coefficient
    """
    sids = list(cal_data.keys())
    for sid in sids:
        if "cameras" in cal_data[sid]["sensor_model"]:
            if not "muzza" in cal_data[sid]["sensor_model"].lower():  # only support opencv model
                focal = cal_data[sid]["focal"]
                pp = cal_data[sid]["pp"]
                K = np.eye(3)
                if isinstance(focal, list):
                    K[0, 0], K[1, 1] = focal
                else:
                    K[0, 0], K[1, 1] = focal, focal
                K[0, 2], K[1, 2] = pp
                cal_data[sid]["K"] = K
                if "inv_poly" in cal_data[sid].keys():
                    cal_data[sid]["D"] = np.array(cal_data[sid]["inv_poly"])
        if "extrinsic" in cal_data[sid].keys():
            tx, ty, tz, qx, qy, qz, qw = cal_data[sid]["extrinsic"]
            T = np.eye(4)
            q = Rotation.from_quat([qx, qy, qz, qw])
            if hasattr(q, "as_dcm"):
                T[:3, :3] = q.as_dcm().astype("float32")
            else:
                T[:3, :3] = q.as_matrix().astype("float32")
            T[:3, 3] = np.array([tx, ty, tz])
            cal_data[sid]["T_es"] = T
            cal_data[sid]["T_se"] = np.linalg.inv(T)


def read_cal_data(yaml_file):
    cal_data = EasyDict()
    with open(yaml_file) as f:
        rig_data = yaml.load(f, Loader=yaml.FullLoader)
        rig_data = EasyDict(rig_data)
        # camera_cal_file_version = rig_data.get('version',1.0)
        cal_data = rig_data.rig
        format_cal_data(cal_data)
        # if camera_cal_file_version<=1.0:
        #     transform_old_rig(cal_data)
    return cal_data


def read_ego_paths(trajectory_path):
    def to_rotation(data):
        t = data[:3]
        R = Rotation.from_quat(data[3:7]).as_matrix()
        T = np.eye(4)
        T[:3, :3] = R
        T[:3, 3] = t
        return T

    a = np.loadtxt(trajectory_path)
    out_dict = {i[0]: to_rotation(i[1:8]) for i in a}
    timestamps = np.array(sorted(out_dict.keys()))
    return out_dict, timestamps


def read_Tel_from_yaml_file(yaml_file):
    cal_data = EasyDict()
    with open(yaml_file) as f:
        rig_data = yaml.load(f, Loader=yaml.FullLoader)
        rig_data = EasyDict(rig_data)
        cal_data = rig_data.rig
    lidar_ext = cal_data['lidar1']['extrinsic']
    tx, ty, tz, qx, qy, qz, qw = lidar_ext
    T = np.eye(4)
    q = Rotation.from_quat([qx, qy, qz, qw])
    if hasattr(q, "as_dcm"):
        T[:3, :3] = q.as_dcm().astype("float32")
    else:
        T[:3, :3] = q.as_matrix().astype("float32")
    T[:3, 3] = np.array([tx, ty, tz])
    return T



from typing import Tuple, Dict
from pathlib import Path
from functools import partial
import yaml
from dataclasses import dataclass, fields

import cv2
import numpy as np
from scipy.optimize import curve_fit
from scipy.spatial.transform import Rotation as R

# from .helpers import find_n_closest_frames

EPS_FLOAT32 = float(np.finfo(np.float32).eps)
MAX_FLOAT32 = float(np.finfo(np.float32).max)


def to_homo(points: np.array) -> np.array:
    if points.ndim == 1:
        points = points[None, :]
    ones = np.ones((len(points), 1), dtype=np.float32)
    return np.concatenate((points, ones), axis=1)


class Calibration:
    def __init__(self, calib_info: dict) -> None:
        self._data = calib_info

    @classmethod
    def from_path(cls, path: Path) -> dict:
        with open(path, "r") as f:
            calib = yaml.safe_load(f)
        return cls(calib)


@dataclass
class Pose:
    timestamp: float
    x: float
    y: float
    z: float
    qx: float
    qy: float
    qz: float
    qw: float

    def __post_init__(self) -> None:
        for field in fields(self):
            setattr(self, field.name, field.type(getattr(self, field.name)))

    @property
    def position(self) -> Tuple[float, float, float]:
        return (self.x, self.y, self.z)

    @property
    def quat(self) -> Tuple[float, float, float, float]:
        return (self.qx, self.qy, self.qz, self.qw)

    @property
    def Rt4x4(self) -> np.ndarray:
        return to_rt4x4(self.position, self.quat)


def to_rt4x4(position: Tuple[float, float, float], quat: Tuple[float, float, float, float]) -> np.ndarray:
    mat4x4 = np.eye(4, dtype=np.float32)
    mat4x4[:3, :3] = R.from_quat(quat).as_matrix()
    mat4x4[:3, 3] = position
    return mat4x4


class Trajectory:
    def __init__(self, poses: Dict[float, Pose]) -> None:
        self.poses = poses

    @classmethod
    def from_path(cls, path: Path) -> "Trajectory":
        poses = {}
        with path.open() as file:
            for line in file:
                timestamp, x, y, z, qx, qy, qz, qw = map(float, line.strip().split())
                pose = Pose(timestamp, x, y, z, qx, qy, qz, qw)
                poses[timestamp] = pose
        return cls(poses)

    def get_pose(self, timestamp: float) -> Pose:
        return self.poses[timestamp]


class FisheyeCameraModel:
    def __init__(self, calib, camera_id):
        camera_calib = calib["rig"][camera_id]
        self.pp = camera_calib["pp"]
        self.focal = camera_calib["focal"]
        self.inv_poly = np.array(camera_calib["inv_poly"], dtype=np.float32)
        self.image_size = camera_calib["image_size"]
        self.fov = camera_calib["fov_fit"] if 'fov_fit' in camera_calib.keys() else 200
        self.intrinsic_mat = np.array(
            [
                [self.focal[0], 0, self.pp[0]],
                [0, self.focal[1], self.pp[1]],
                [0, 0, 1],
            ],
            dtype=np.float32,
        )

    @classmethod
    def from_intrinsic_array(cls, intrinsic, camera_id):
        """fisheye camera; usually the intrinsic array contains 8 number"""
        calib = {"rig": {
            camera_id: {
                "pp": intrinsic[:2],
                "focal": intrinsic[2:4],
                "inv_poly": intrinsic[4:8],
                "image_size": [1920, 1080],
                "fov_fit": 200,
            }
        }}
        return FisheyeCameraModel(calib, camera_id)

    @staticmethod
    def fit_unproj_func(p0, p1, p2, p3, fov=200):
        def proj_func(x, params):
            p0, p1, p2, p3 = params
            return x + p0 * x ** 3 + p1 * x ** 5 + p2 * x ** 7 + p3 * x ** 9

        def poly_odd6(x, k0, k1, k2, k3, k4, k5):
            return x + k0 * x ** 3 + k1 * x ** 5 + k2 * x ** 7 + k3 * x ** 9 + k4 * x ** 11 + k5 * x ** 13

        theta = np.linspace(-0.5 * fov * np.pi / 180, 0.5 * fov * np.pi / 180, 2000)
        theta_d = proj_func(theta, (p0, p1, p2, p3))
        params, pcov = curve_fit(poly_odd6, theta_d, theta)
        error = np.sqrt(np.diag(pcov)).mean()
        assert error < 1e-3, "poly parameter curve fitting failed: {:f}.".format(error)
        k0, k1, k2, k3, k4, k5 = params
        return partial(poly_odd6, k0=k0, k1=k1, k2=k2, k3=k3, k4=k4, k5=k5)

    def unproject_points(self, points):
        unproj_func = self.fit_unproj_func(*self.inv_poly[:4])
        cx, cy = self.pp
        fx, fy = self.focal
        u = points[:, 0]
        v = points[:, 1]
        x_distorted = (u - cx) / fx
        y_distorted = (v - cy) / fy
        r_distorted = theta_distorted = np.sqrt(x_distorted ** 2 + y_distorted ** 2)
        r_distorted[r_distorted < 1e-5] = 1e-5
        theta = unproj_func(r_distorted)
        theta = np.clip(theta, - 0.5 * self.fov * np.pi / 180, 0.5 * self.fov * np.pi / 180)
        vignette_mask = np.float32(np.abs(theta * 180 / np.pi) < self.fov / 2)
        # camera coords on a sphere x-y-z right-down-forward
        dd = np.sin(theta)
        xx = x_distorted * dd / r_distorted
        yy = y_distorted * dd / r_distorted
        zz = np.cos(theta)
        fisheye_cam_coords = np.stack([xx, yy, zz], axis=1)
        return fisheye_cam_coords

    def project_points(self, points: np.ndarray) -> np.ndarray:
        """_summary_

        Args:
            points (np.ndarray): n*3 np

        Returns:
            np.ndarray: n*2 np
        """
        # 涉及到的参数，1 fov 2 inv_poly 3 intrinsic: focal /pp
        xc = points[:, 0]
        yc = points[:, 1]
        zc = points[:, 2]
        norm = np.sqrt(xc ** 2 + yc ** 2)
        theta = np.arctan2(norm, zc)
        fov_mask = theta > self.fov / 2 * np.pi / 180
        rho = (
                theta
                + self.inv_poly[0] * theta ** 3
                + self.inv_poly[1] * theta ** 5
                + self.inv_poly[2] * theta ** 7
                + self.inv_poly[3] * theta ** 9
        )
        width, height = self.image_size
        image_radius = np.sqrt((width / 2) ** 2 + (height) ** 2)
        rho[fov_mask] = 2 * image_radius / self.focal[0]
        xn = rho * xc / norm
        yn = rho * yc / norm
        xn[norm < EPS_FLOAT32] = 0
        yn[norm < EPS_FLOAT32] = 0
        norm_coords = np.stack([xn, yn, np.ones_like(xn)], axis=1)
        intrinsic_mat = np.array(
            [
                [self.focal[0], 0, self.pp[0]],
                [0, self.focal[1], self.pp[1]],
                [0, 0, 1],
            ],
            dtype=np.float32,
        )

        image_coords = norm_coords @ intrinsic_mat.T
        return image_coords[:, :2]

    def undistort_image(self, image: np.ndarray) -> np.ndarray:
        return cv2.fisheye.undistortImage(
            distorted=image,
            K=self.intrinsic_mat,
            D=self.inv_poly[:4],
            Knew=self.intrinsic_mat
        )

    def undistort_points(self, points: np.ndarray, out_as_homo=True) -> np.ndarray:
        """Undistort 2D points on distorted image to 2D points on undistorted image."""
        norm_coords_homo = self.unproject_points(points)
        undistorted_points_on_image = norm_coords_homo @ self.intrinsic_mat.T  # 用小孔相机（假设无畸变）的方式将相机系中的3D点投到图像上
        if not out_as_homo:
            undistorted_points_on_image = undistorted_points_on_image[:, :2] / undistorted_points_on_image[:, 2:]
        return undistorted_points_on_image

    # def unproject_points(self, points: np.ndarray) -> np.ndarray:
    #     """Undistort 2D points on distorted image to 3D points in camera coordinate system (with Z guessed, Z=1)."""
    #     norm_coords = cv2.undistortPoints(
    #         points.reshape(1, -1, 2),
    #         self.intrinsic_mat,
    #         self.inv_poly,
    #     )
    #     return np.squeeze(cv2.convertPointsToHomogeneous(norm_coords))


def get_camera_models(calib_path, cameras=None):
    calib = Calibration.from_path(calib_path)
    cameras = [s for s in calib._data['rig'].keys() if 'cam' in s] if cameras is None else cameras
    camera_models = {cam_id: FisheyeCameraModel(calib._data, cam_id) for cam_id in cameras}
    return camera_models


def get_camera_models_from_intrinsic_array(array_dict, cameras=None):
    cameras = list(array_dict.keys()) if cameras is None else cameras
    camera_models = {cam_id: FisheyeCameraModel.from_intrinsic_array(array_dict[cam_id]) for cam_id in cameras}
    return camera_models


def to_camera_xy(calib_path, cam_id, points):
    camera = FisheyeCameraModel(Calibration.from_path(calib_path)._data, cam_id)
    return camera.project_points(points)



import numpy as np
from scipy.spatial.transform import Rotation
import sys

sys.path.append(".")
import copy

"""
所有的bbx的形式
前5个都是前左上
1 collaborator
2 psr
3 4djson
4 7points
5 9points
6 others: pcddet box 
"""


def jsonbox_to_box9d(box):
    p = box['pos_xyz']
    s = box['scale_xyz']
    r = box['rot_xyz']
    return list(p) + list(s) + list(r)


def lfcjson_to_box9d(box):
    p = box['translation']
    s = box['size']
    r = Rotation.from_quat(box['rotation']).as_euler("XYZ")
    return list(p) + list(s) + list(r)


def fbbox_to_box9d(box):
    p = box['translation']
    s = box['size']
    r = Rotation.from_matrix(box['rotation'], scalar_first=True).as_euler("XYZ")
    return list(p) + list(s) + list(r)


def to_corners_7(pts):
    """

    lidar cor:
       2 -------- 3
      /|         /|
    6 -------- 7 .
    | |        | |
    . 1 -------- 0
    |/         |/
    5 -------- 4

    """
    x, y, z = pts[:3]
    a, b, c = pts[3:6] / 2
    u, v, w = 0, 0, pts[6]
    corners = np.array(
        [
            [+a, -b, -c],
            [+a, +b, -c],
            [+a, +b, +c],
            [+a, -b, +c],
            [-a, -b, -c],
            [-a, +b, -c],
            [-a, +b, +c],
            [-a, -b, +c],
        ]
    )
    R = Rotation.from_euler("XYZ", [u, v, w]).as_matrix()
    corners = (R @ corners.T).T
    corners += np.array([x, y, z]).reshape(-1, 3)
    return corners


def to_corners_9(pts):
    """

    lidar cor:
       2 -------- 3
      /|         /|
    6 -------- 7 .
    | |        | |
    . 1 -------- 0
    |/         |/
    5 -------- 4

    """
    pts = np.array(pts)
    x, y, z = pts[:3]
    a, b, c = pts[3:6] / 2
    u, v, w = pts[6], pts[7], pts[8]
    corners = np.array(
        [
            [+a, -b, -c],
            [+a, +b, -c],
            [+a, +b, +c],
            [+a, -b, +c],
            [-a, -b, -c],
            [-a, +b, -c],
            [-a, +b, +c],
            [-a, -b, +c],
        ]
    )

    R = Rotation.from_euler("XYZ", [u, v, w]).as_matrix()
    corners = (R @ corners.T).T
    corners += np.array([x, y, z]).reshape(-1, 3)
    return corners


def anno_box_to_9_values_box(box: dict) -> np.ndarray:
    return np.array(
        [
            box["psr"]["position"]["x"],
            box["psr"]["position"]["y"],
            box["psr"]["position"]["z"],
            box["psr"]["scale"]["x"],
            box["psr"]["scale"]["y"],
            box["psr"]["scale"]["z"],
            box["psr"]["rotation"]["x"],
            box["psr"]["rotation"]["y"],
            box["psr"]["rotation"]["z"],
        ],
        dtype=np.float32,
    )


def transform_output_geometry_to_array(box):
    return [
        *box["pos_xyz"],
        *box["scale_xyz"],
        *box["rot_xyz"],
    ]


def transform_psr_to_array(psr):
    return [
        psr["position"]["x"],
        psr["position"]["y"],
        psr["position"]["z"],
        psr["scale"]["x"],
        psr["scale"]["y"],
        psr["scale"]["z"],
        psr["rotation"]["x"],
        psr["rotation"]["y"],
        psr["rotation"]["z"],
    ]


def translate_psr_to_output_geometry(psr):
    output = {}
    output["pos_xyz"] = [psr["position"]["x"], psr["position"]["y"], psr["position"]["z"]]
    output["scale_xyz"] = [psr["scale"]["x"], psr["scale"]["y"], psr["scale"]["z"]]
    output["rot_xyz"] = [psr["rotation"]["x"], psr["rotation"]["y"], psr["rotation"]["z"]]
    return output


def translate_output_geometry_to_psr(box):
    output = {
        "position": {
            "x": box["pos_xyz"][0],
            "y": box["pos_xyz"][1],
            "z": box["pos_xyz"][2],
        },
        "scale": {
            "x": box["scale_xyz"][0],
            "y": box["scale_xyz"][1],
            "z": box["scale_xyz"][2],
        },
        "rotation": {
            "x": box["rot_xyz"][0],
            "y": box["rot_xyz"][1],
            "z": box["rot_xyz"][2],
        },
    }
    return output


def to_list_psr_box_from_vec7_conf(boxes, confs, labels):
    output = []
    for b, c, l in zip(boxes, confs, labels):
        output.append({"obj_attr": {}, "obj_type": l, "psr": box_vec7_to_psr(b), "conf": c})
    return output


def box_vec7_to_psr(vec7):
    return {
        "position": {
            "x": vec7[0],
            "y": vec7[1],
            "z": vec7[2],
        },
        "scale": {
            "x": vec7[3],
            "y": vec7[4],
            "z": vec7[5],
        },
        "rotation": {"x": 0, "y": 0, "z": vec7[6]},
    }


def box_vec9_to_psr(vec9):
    return {
        "position": {
            "x": vec9[0],
            "y": vec9[1],
            "z": vec9[2],
        },
        "scale": {
            "x": vec9[3],
            "y": vec9[4],
            "z": vec9[5],
        },
        "rotation": {
            "x": vec9[6],
            "y": vec9[7],
            "z": vec9[8]},
    }


def box_vec7_to_Ts(vec7):
    p = vec7[:3]
    s = vec7[3:6]
    r = vec7[6]
    T = np.eye(4)
    T[:3, :3] = Rotation.from_euler('z', r).as_matrix()
    T[:3, 3] = p
    return T, s


def box_Ts_to_vec9(T, s):
    return T[:3, 3].tolist() + list(s) + Rotation.from_matrix(T[:3, :3]).as_euler('XYZ').tolist()


def frame_box_to_psr(box):
    return {
        "position": {
            "x": box['pos_xyz'][0],
            "y": box['pos_xyz'][1],
            "z": box['pos_xyz'][2],
        },
        "scale": {
            "x": box['scale_xyz'][0],
            "y": box['scale_xyz'][1],
            "z": box['scale_xyz'][2],
        },
        "rotation": {
            "x": box['rot_xyz'][0],
            "y": box['rot_xyz'][1],
            "z": box['rot_xyz'][2],
        },
    }


def translate_psr_with_T(psr, Tts):
    # Tts 表示 从box的当前系转到目标系
    position = np.array([[psr["position"]["x"], psr["position"]["y"], psr["position"]["z"], 1.0]])
    rotation = np.eye(4)
    rotation[:3, :3] = Rotation.from_euler("XYZ", [psr["rotation"]["x"], psr["rotation"]["y"],
                                                   psr["rotation"]["z"]]).as_matrix()
    transformed_position = (position @ Tts.T).squeeze()
    transformed_rotation = Rotation.from_matrix((Tts @ rotation)[:3, :3]).as_euler("XYZ")
    new_psr = copy.deepcopy(psr)
    new_psr["position"] = {
        "x": transformed_position[0],
        "y": transformed_position[1],
        "z": transformed_position[2],
    }
    new_psr["rotation"] = {
        "x": transformed_rotation[0],
        "y": transformed_rotation[1],
        "z": transformed_rotation[2],
    }
    return new_psr


def box_corners_to_dot_cloud(corners):
    # 保证是3维度的点
    assert np.array(corners).shape[-1] == 3

    def get_lines(box):
        output = []
        for i, j in [
            [box[0], box[1]],
            [box[1], box[2]],
            [box[2], box[3]],
            [box[3], box[0]],
            [box[0 + 4], box[1 + 4]],
            [box[1 + 4], box[2 + 4]],
            [box[2 + 4], box[3 + 4]],
            [box[3 + 4], box[0 + 4]],
            [box[0], box[0 + 4]],
            [box[1], box[1 + 4]],
            [box[2], box[2 + 4]],
            [box[3], box[3 + 4]],
        ]:
            for n in range(100):
                n = n / 100
                output += [i * (1 - n) + j * n]
        return output

    box_points = np.array([get_lines(i) for i in corners.reshape(-1, 8, 3)]).reshape(-1, 3)
    return box_points



import numpy as np
from scipy.spatial.transform import Rotation


def to_homo(points: np.array) -> np.array:
    if points.ndim == 1:
        points = points[None, :]
    ones = np.ones((len(points), 1), dtype=np.float32)
    return np.concatenate((points, ones), axis=1)


def transform_pts_with_T(points, T):
    # from pts3d to lidar 3d
    points = np.array(points)
    shape = points.shape
    points = points.reshape(-1, 3)
    points_output = (to_homo(points) @ T.T)[:, :3].reshape(*shape)
    return points_output

def Rt2T(R, t):
    T = np.eye(4)
    T[:3, :3] = R
    T[:3, 3] = t
    return T

def T2Rt(T):
    return T[:3, :3], T[:3, 3]

def rotation_to_wxyz(rot):
    R = np.array(rot).reshape(3, 3)
    return Rotation.from_matrix(R).as_quat(scalar_first=True).tolist()  # wxyz, default False

def rotation_to_xyzw(rot):
    R = np.array(rot).reshape(3, 3)
    return Rotation.from_matrix(R).as_quat(scalar_first=False).tolist()



import copy
import torch
from tqdm import tqdm
import torch.multiprocessing as torch_mp
import json
from pathlib import Path as P
import numpy as np
from typing import List, Union
import yaml
from easydict import EasyDict
from scipy.spatial.transform import Rotation
import time
import os
import re
import os.path as op
import subprocess
from pathlib import Path
import multiprocessing as mp
EPS_FLOAT32 = float(np.finfo(np.float32).eps)
MAX_FLOAT32 = float(np.finfo(np.float32).max)

ipm_cls_dict = {
    0: "lanemark",
    1: "parkingline",
    2: "deceleraion_hump",
    3: "warningpole",
    4: "warningcone",
    5: "guideline",
    6: "groundpin",
    7: "groundlock",
    8: "zebracrosswalk",
}

label_to_class = {
    "101": "Box_truck",
    "102": "Truck",
    "103": "Car",
    "104": "Van",
    "105": "Bus",
    "106": "Engineering_vehicle",
    "201": "Pedestrian",
    "202": "Cyclist",
    "301": "Bicycle",
    "100": "DontCare",
}
velocitys_threshold_dict = {
    "Box_truck": 1,
    "Car": 1,
    "Pedestrian": 0.5,
    "Cyclist": 0.5,
    "Van": 1,
    "Truck": 1,
    "Bus": 1,
    "Engineering_vehicle": 1,
    "Bicycle": 0.5,
    "DontCare": 10,
}  # m/s

rotate_class = ["Box_truck", "Van", "Car", "Truck", "Bus", "Cyclist", "Bicycle", "Engineering_vehicle"]

min_l_dict = {
    "Box_truck": 3,
    "Car": 3,
    "Pedestrian": 0.1,
    "Cyclist": 1,
    "Van": 3,
    "Truck": 3,
    "Bus": 3,
    "Engineering_vehicle": 3,
    "Bicycle": 1,
    "DontCare": 0.1,
}  # m

min_points_dict = {
    "Box_truck": 10,
    "Car": 10,
    "Pedestrian": 5,
    "Cyclist": 5,
    "Van": 10,
    "Truck": 10,
    "Bus": 10,
    "Engineering_vehicle": 10,
    "Bicycle": 5,
    "DontCare": 5,
}

side_cameras = ["sv_left", "sv_right"]
main_cameras = ["sv_front", "sv_rear"]

sensor_time_offset = {"camera_sv_front": -50, "camera_sv_left": -50, "camera_sv_right": -50, "camera_sv_rear": -50}


def generate_lidar_mask(points, img_shape, CameraExtrinsicMat, CameraMat, DistCoeff, filter_xy=True, use_gpu=True, fov=200, round_pixel=True):
    """
    only support opencv camera;  points_raw: lidar系， img_shape, sensor2lidar，camera_K, D
    """
    is_numpy = False
    if isinstance(points, np.ndarray):
        points = torch.from_numpy(points).float()
        is_numpy = True
    if isinstance(img_shape, np.ndarray):
        img_shape = torch.from_numpy(img_shape).float()
    elif isinstance(img_shape, list) or isinstance(img_shape, tuple):
        img_shape = torch.from_numpy(np.array(img_shape)).float()
    if isinstance(CameraExtrinsicMat, np.ndarray):
        CameraExtrinsicMat = torch.from_numpy(CameraExtrinsicMat).float()
    if isinstance(CameraMat, np.ndarray):
        CameraMat = torch.from_numpy(CameraMat).float()
    if isinstance(DistCoeff, np.ndarray):
        DistCoeff = torch.from_numpy(DistCoeff).float()
    elif isinstance(DistCoeff, list):
        DistCoeff = torch.from_numpy(np.array(DistCoeff)).float()
    device = points.device
    if use_gpu:
        points = points.cuda()
        img_shape = img_shape.cuda()
        CameraExtrinsicMat = CameraExtrinsicMat.cuda()
        CameraMat = CameraMat.cuda()
        DistCoeff = DistCoeff.cuda()
    assert len(DistCoeff) in [4, 5]
    p = points[:, :3]
    # p=np.hstack((p,np.zeros((p.shape[0],1))+1))
    p = torch.cat([p, torch.ones((p.shape[0], 1), device=p.device)], dim=1)
    # camera_coordinates=np.dot(CameraExtrinsicMat,p.T)
    camera_coordinates = torch.matmul(CameraExtrinsicMat, p.T)
    # index = np.argsort(-camera_coordinates[2,:])
    # camera_coordinates = camera_coordinates[:,index]
    # front_filter = camera_coordinates[2,:]>0
    # camera_coordinates=camera_coordinates[:,camera_coordinates[2,:]>0]
    camera_coords = camera_coordinates.T
    if len(DistCoeff) == 5:
        fov_mask = camera_coordinates[2, :] < 0
        normalize_coordinates = camera_coordinates / (camera_coordinates[2, :] + EPS_FLOAT32)
        xc = copy.deepcopy(normalize_coordinates[0, :])
        yc = copy.deepcopy(normalize_coordinates[1, :])
        r2 = xc * xc + yc * yc
        tmpdist = 1 + DistCoeff[0] * r2 + DistCoeff[1] * r2 * r2 + DistCoeff[4] * r2 * r2 * r2
        normalize_coordinates[0, :] = xc * tmpdist + 2 * DistCoeff[2] * xc * yc + DistCoeff[3] * (r2 + 2 * xc * xc)
        normalize_coordinates[1, :] = yc * tmpdist + DistCoeff[2] * (r2 + 2 * yc * yc) + 2 * DistCoeff[3] * xc * yc
    elif len(DistCoeff) == 4:
        xc = copy.deepcopy(camera_coordinates[0, :])
        yc = copy.deepcopy(camera_coordinates[1, :])
        zc = copy.deepcopy(camera_coordinates[2, :])
        r2 = xc * xc + yc * yc
        norm = torch.sqrt(r2)
        theta = torch.atan2(norm, zc)
        FOV = fov
        FOV_ = FOV / 2 * np.pi / 180
        fov_mask = theta > FOV_
        rho = theta + DistCoeff[0] * theta**3 + DistCoeff[1] * theta**5 + DistCoeff[2] * theta**7 + DistCoeff[3] * theta**9
        image_radius = torch.sqrt((img_shape[1] / 2) ** 2 + (img_shape[0]) ** 2)
        focal = [CameraMat[0, 0], CameraMat[1, 1]]
        # if filter_xy:
        rho[fov_mask] = 2 * image_radius / focal[0]
        # else:
        #     rho[fov_mask] = FOV_ + DistCoeff[0] * FOV_**3 + DistCoeff[1] * FOV_**5 + \
        #                 DistCoeff[2] * FOV_**7 + DistCoeff[3] * FOV_**9
        xn = rho * xc / norm
        yn = rho * yc / norm
        xn[norm < EPS_FLOAT32] = 0
        yn[norm < EPS_FLOAT32] = 0
        # normalize_coordinates = np.concatenate([np.vstack([xn, yn]), np.ones((1,xn.shape[0]))],axis=0)
        normalize_coordinates = torch.cat([torch.vstack([xn, yn]), torch.ones((1, xn.shape[0]), device=xn.device)], dim=0)

    # pixel_coordinates=np.dot(CameraMat[:3,:3],normalize_coordinates)
    pixel_coordinates = torch.matmul(CameraMat[:3, :3], normalize_coordinates)
    # z_filter=pixel_coordinates[2,:]>0
    x_filter = torch.logical_and(pixel_coordinates[0, :] > 0, pixel_coordinates[0, :] < img_shape[1])
    y_filter = torch.logical_and(pixel_coordinates[1, :] > 0, pixel_coordinates[1, :] < img_shape[0])
    all_filter = x_filter & y_filter & ~fov_mask
    if filter_xy:
        pixel_coordinates = pixel_coordinates[:, all_filter]
        camera_coords = camera_coords[all_filter]
        # if len(DistCoeff)==5:
        #     all_filter = x_filter & y_filter & ~fov_mask
        # elif len(DistCoeff)==4:
        #     all_filter = x_filter & y_filter & ~fov_mask
    # else:
    # if len(DistCoeff)==5:
    #     all_filter = torch.ones_like(z_filter).bool()
    # elif len(DistCoeff)==4:
    #     all_filter = torch.ones_like(z_filter).bool()
    # all_filter = torch.ones_like(fov_mask).bool()
    pixel_coordinates = pixel_coordinates.T
    if round_pixel:
        pixel_coordinates = pixel_coordinates.long()
    img_shape = img_shape.long().cpu().numpy()
    mask = torch.zeros((img_shape[0], img_shape[1], 3), device=camera_coords.device)
    if filter_xy:
        mask[pixel_coordinates[:, 1], pixel_coordinates[:, 0]] = camera_coords[:, :3]
        if is_numpy:
            return (mask.cpu().numpy(), all_filter.cpu().numpy(), pixel_coordinates.cpu().numpy())
        else:
            return (mask.to(device), all_filter.to(device), pixel_coordinates.to(device))
    else:
        if is_numpy:
            return (all_filter.cpu().numpy(), pixel_coordinates.cpu().numpy())
        else:
            return (all_filter.to(device), pixel_coordinates.to(device))


def format_cal_data(cal_data):
    """
    generate extrinsic mat, camera intrinsic mat and distort coefficient
    """
    sids = list(cal_data.keys())
    # if 'cam_bev' in sids:
    #     sids.remove('cam_bev')
    for sid in sids:
        if "cameras" in cal_data[sid]["sensor_model"]:
            if not "muzza" in cal_data[sid]["sensor_model"].lower():  # only support opencv model
                focal = cal_data[sid]["focal"]
                pp = cal_data[sid]["pp"]
                K = np.eye(3)
                if isinstance(focal, list):
                    K[0, 0], K[1, 1] = focal
                else:
                    K[0, 0], K[1, 1] = focal, focal
                K[0, 2], K[1, 2] = pp
                cal_data[sid]["K"] = K
                if "inv_poly" in cal_data[sid].keys():
                    cal_data[sid]["D"] = np.array(cal_data[sid]["inv_poly"])
        if "extrinsic" in cal_data[sid].keys():
            tx, ty, tz, qx, qy, qz, qw = cal_data[sid]["extrinsic"]
            T = np.eye(4)
            q = Rotation.from_quat([qx, qy, qz, qw])
            if hasattr(q, "as_dcm"):
                T[:3, :3] = q.as_dcm().astype("float32")
            else:
                T[:3, :3] = q.as_matrix().astype("float32")
            T[:3, 3] = np.array([tx, ty, tz])
            cal_data[sid]["T_es"] = T
            cal_data[sid]["T_se"] = np.linalg.inv(T)


def read_cal_data(yaml_file):
    cal_data = EasyDict()
    with open(yaml_file) as f:
        rig_data = yaml.load(f, Loader=yaml.FullLoader)
        rig_data = EasyDict(rig_data)
        # camera_cal_file_version = rig_data.get('version',1.0)
        cal_data = rig_data.rig
        format_cal_data(cal_data)
        # if camera_cal_file_version<=1.0:
        #     transform_old_rig(cal_data)
    return cal_data


def get_sync_filename(file_names, file_times, sync_time, return_idx=False):
    """
    get sync filename in file_root according to the sync_time
    """
    times_diff = np.abs(file_times - sync_time)
    min_index = np.argmin(times_diff)
    if times_diff[min_index] < 15:
        sync_file_name = file_names[min_index]
        if return_idx:
            return sync_file_name, min_index
        return sync_file_name
    else:
        if return_idx:
            return None, None
        return None


def get_times(file_root, sync_offset=None, no_sync_list=None, return_frame=False, tag=None):
    if sync_offset is not None:
        id = op.basename(file_root)
    file_names = os.listdir(file_root)
    times = []
    file_names = [x for x in file_names if not "html" in x]
    if tag is not None:
        file_names = [x for x in file_names if tag in x]
    frames = []
    for file_name in file_names:
        if "html" in file_name:
            time = 0
        elif "indices" in file_name:
            time = 0
        elif "_" in file_name:
            time = float(file_name.split("_")[1].split(".")[0])
            frames.append(int(file_name.split("_")[0]))
        else:
            time = float(file_name.split(".")[0])
        if sync_offset is not None:
            assert no_sync_list is not None, "must input no_sync_list"
            if not id in no_sync_list:
                time -= sync_offset
        times.append(time)
    times = np.array(times)
    file_names = np.array(file_names)
    if len(frames) > 0:
        frames = np.array(frames)
        sort_ind = np.argsort(frames)
        frames = frames[sort_ind]
    else:
        sort_ind = np.argsort(times)
    times = times[sort_ind]
    file_names = file_names[sort_ind]
    if return_frame and len(frames) > 0:
        return times, file_names, frames
    else:
        return times, file_names


def get_round_timestamp(timestamps, filter_unsync=False):
    round_timestamps = (np.round(timestamps / 100) * 100).astype(np.int64)
    offsets = abs(round_timestamps - timestamps)
    valid_mask = offsets < 15
    valid_round_timestamps = round_timestamps[valid_mask]
    if filter_unsync:
        return valid_round_timestamps, valid_mask
    else:
        return round_timestamps


def anno_box_to_7_values_box(box: dict) -> np.ndarray:
    return np.array(
        [
            [
                box["psr"]["position"]["x"],
                box["psr"]["position"]["y"],
                box["psr"]["position"]["z"],
                box["psr"]["scale"]["x"],
                box["psr"]["scale"]["y"],
                box["psr"]["scale"]["z"],
                box["psr"]["rotation"]["z"],
            ]
        ],
        dtype=np.float32,
    )

def anno_box_to_9_values_box(box: dict) -> np.ndarray:
    return np.array(
        [
            [
                box["psr"]["position"]["x"],
                box["psr"]["position"]["y"],
                box["psr"]["position"]["z"],
                box["psr"]["scale"]["x"],
                box["psr"]["scale"]["y"],
                box["psr"]["scale"]["z"],
                box["psr"]["rotation"]["x"],
                box["psr"]["rotation"]["y"],
                box["psr"]["rotation"]["z"],
            ]
        ],
        dtype=np.float32,
    )

def read_json_to_list(path):
    with open(str(path)) as f:
        a = json.load(f)
    return a
def read_json(path):
    with open(str(path)) as f:
        a = json.load(f)
    return a


def format_floats(obj):
    if isinstance(obj, float):
        return round(obj, 4)
    elif isinstance(obj, dict):
        for k, v in obj.items():
            obj[k] = format_floats(v)
        return obj
    elif isinstance(obj, list):
        for i, v in enumerate(obj):
            obj[i] = format_floats(v)
        return obj
    else:
        return obj


def write_json_from_list(output_dict, save_path, format_float=False, indent=None):
    save_path = str(save_path)
    P(save_path).parent.mkdir(exist_ok=True, parents=True)
    if format_float:
        output_dict = format_floats(output_dict)
    with open(save_path, "w") as f:
        # json.dump(output_dict, f, indent=4, default=lambda x: round(x, 3))
        json.dump(output_dict, f, default=lambda x: round(x, 3), indent=indent)


def transform_points_ego_to_lidar(points: np.ndarray, ego_to_lidar: np.ndarray) -> np.ndarray:
    assert points.ndim == 2
    if points.shape[1] == 3:
        points = np.concatenate([points, np.ones(points.shape[0])[:, None]], axis=1)
    camera_coords = points @ ego_to_lidar.T
    return camera_coords[:, :3]

class CameraParam:
    def __init__(self, calib: dict, middle_lidar_id: str, camera_id: str) -> None:
        self.lidar_to_ego = calib[middle_lidar_id]["T_es"]
        self.ego_to_lidar = np.linalg.inv(self.lidar_to_ego)
        self.cam_to_ego = calib[camera_id]["T_es"]
        self.ego_to_cam = np.linalg.inv(self.cam_to_ego)
        self.lidar_to_cam = np.dot(np.linalg.inv(self.cam_to_ego), self.lidar_to_ego)
        self.cam_to_lidar = np.linalg.inv(self.lidar_to_cam)
        self.image_shape = calib[camera_id]["image_size"][::-1]
        self.K = calib[camera_id]["K"]
        self.D = calib[camera_id]["D"]
        self.camera_id = camera_id
        self.fov = calib[camera_id]['fov_fit']
        self.inv_poly = calib[camera_id]['inv_poly']
        self.image_width, self.image_height = calib[camera_id]['image_size']
        self.image_radius = np.sqrt((self.image_width / 2) ** 2 + (self.image_height) ** 2)

def check_whether_points_behind_camera(points: torch.Tensor, ego_to_cam: torch.Tensor) -> torch.Tensor:
    if points.ndim == 2 and points.shape[1] == 3:
        points = torch.cat([points, torch.ones((points.shape[0], 1), device=points.device)], dim=1)
    camera_coords = points @ ego_to_cam.T
    return camera_coords[:, 2] < 0


def find_path_from_ts_and_dir_path(ts, dir_path):
    times, file_names = get_times(dir_path)
    return op.join(dir_path, get_sync_filename(file_names, times, ts))


def find_path_from_ts_and_dir_path_maybe_not_find(ts, dir_path):
    times, file_names = get_times(dir_path)
    if get_sync_filename(file_names, times, ts) is None:
        return None
    return op.join(dir_path, get_sync_filename(file_names, times, ts))



def read_ts_json(path):
    ts_dict_list = read_json_to_list(path)
    output_dict = {i["lidar"]: i for i in ts_dict_list}
    ts_list = sorted([i["lidar"] for i in ts_dict_list])
    return ts_list, output_dict


def torch_pool(func, list_to_process, n_processes=torch_mp.cpu_count()):
    torch_mp.set_start_method("spawn", force=True)
    with  torch_mp.Pool(n_processes) as pool:
        r = list(tqdm(pool.imap(func, list_to_process), total=len(list_to_process)))
    return r

def mp_pool(func, list_to_process, n_processes=mp.cpu_count()):
    with mp.Pool(n_processes) as p:
        r = list(tqdm(p.imap(func, list_to_process), total=len(list_to_process)))
    return r



import os
from time import time
from collections import defaultdict
import torch.multiprocessing as torch_mp
import multiprocessing as mp
from tqdm import tqdm
import os.path as op
import numpy as np


class Time:
    def __init__(self):
        self.time = time()

    def tic(self):
        self.time = time()

    def toc(self, info=None):
        if info is not None:
            print(time() - self.time, info)
        else:
            print(time() - self.time)
        self.time = time()


def defaultdict_lambda():
    return defaultdict(dict)


def torch_pool(func, list_to_process, n_processes=torch_mp.cpu_count()):
    torch_mp.set_start_method("spawn", force=True)
    with torch_mp.Pool(n_processes) as pool:
        r = list(tqdm(pool.imap(func, list_to_process), total=len(list_to_process)))
    return r


def mp_pool(func, list_to_process, n_processes=mp.cpu_count()):
    with mp.Pool(n_processes) as p:
        r = list(tqdm(p.imap(func, list_to_process), total=len(list_to_process)))
    return r


def get_sync_filename(file_names, file_times, sync_time, return_idx=False):
    """
    get sync filename in file_root according to the sync_time
    """
    times_diff = np.abs(file_times - sync_time)
    min_index = np.argmin(times_diff)
    if times_diff[min_index] < 15:
        sync_file_name = file_names[min_index]
        if return_idx:
            return sync_file_name, min_index
        return sync_file_name
    else:
        if return_idx:
            return None, None
        return None


def get_times(file_root, sync_offset=None, no_sync_list=None, return_frame=False, tag=None):
    if sync_offset is not None:
        id = op.basename(file_root)
    file_names = os.listdir(file_root)
    times = []
    file_names = [x for x in file_names if not "html" in x]
    if tag is not None:
        file_names = [x for x in file_names if tag in x]
    frames = []
    for file_name in file_names:
        if "html" in file_name:
            time = 0
        elif "indices" in file_name:
            time = 0
        elif "_" in file_name:
            time = float(file_name.split("_")[1].split(".")[0])
            frames.append(int(file_name.split("_")[0]))
        else:
            time = float(file_name.split(".")[0])
        if sync_offset is not None:
            assert no_sync_list is not None, "must input no_sync_list"
            if not id in no_sync_list:
                time -= sync_offset
        times.append(time)
    times = np.array(times)
    file_names = np.array(file_names)
    if len(frames) > 0:
        frames = np.array(frames)
        sort_ind = np.argsort(frames)
        frames = frames[sort_ind]
    else:
        sort_ind = np.argsort(times)
    times = times[sort_ind]
    file_names = file_names[sort_ind]
    if return_frame and len(frames) > 0:
        return times, file_names, frames
    else:
        return times, file_names


def get_round_timestamp(timestamps, filter_unsync=False):
    round_timestamps = (np.round(timestamps / 100) * 100).astype(np.int64)
    offsets = abs(round_timestamps - timestamps)
    valid_mask = offsets < 15
    valid_round_timestamps = round_timestamps[valid_mask]
    if filter_unsync:
        return valid_round_timestamps, valid_mask
    else:
        return round_timestamps


def find_path_from_ts_and_dir_path(ts, dir_path):
    times, file_names = get_times(dir_path)
    return op.join(dir_path, get_sync_filename(file_names, times, ts))


def find_path_from_ts_and_dir_path_maybe_not_find(ts, dir_path):
    times, file_names = get_times(dir_path)
    if get_sync_filename(file_names, times, ts) is None:
        return None
    return op.join(dir_path, get_sync_filename(file_names, times, ts))



import open3d as o3d
import numpy as np
from pathlib import Path as P
import os.path as op
from mtv4d.utils.io_base import read_points_from_bin, write_bin

from scipy.spatial import Delaunay
from pathlib import Path as P
import scipy


def in_hull(p, hull):
    """
    :param p: (N, K) test points
    :param hull: (M, K) M corners of a box
    :return (N) bool
    """
    try:
        if not isinstance(hull, Delaunay):
            hull = Delaunay(hull)
        flag = hull.find_simplex(p) >= 0
    except scipy.spatial.qhull.QhullError:
        print('Warning: not a hull %s' % str(hull))
        flag = np.zeros(p.shape[0], dtype=np.bool)
    return flag


def hpr_index(points, radius=None, view_point=None):
    radius = radius or np.abs(points[:, :3]).max() * 100
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(points[:, :3])
    viewpoint = np.array([0, 0, 0]) if view_point is None else np.array(view_point)
    _, pt_map = pcd.hidden_point_removal(viewpoint, radius)
    return pt_map


def open3d_hpr(points, radius=3000.0, view_point=None):
    pt_map = hpr_index(points[:, :3], radius=radius, view_point=view_point)
    return points[np.array(sorted(pt_map))]


def hidden_point_generate(path):
    root_path = P(path.parent.parent.parent)
    points = read_points_from_bin(str(path))
    points = open3d_hpr(points)
    save_path = op.join(root_path, "hidden_point_removal/lidar1", path.name)
    write_bin(points, save_path)


def pick_visible_points_according_to_hpr_points(poly_points, hpr_points, thres=0.2):
    import torch
    distances = torch.cdist(torch.tensor(poly_points).double(), torch.tensor(hpr_points).double())
    idx = torch.argsort(distances, axis=1)
    dist_smallest = torch.take_along_dim(distances, idx[:, :1], axis=1)[:, 0]
    # dist_smallest = distances[:, idx[:, 0]]
    a = dist_smallest < thres
    return "".join(["1" if i else "0" for i in a])


def draw_points(points, save_path='output.ply'):
    # need points
    points = points[:, :3]
    add_points = np.zeros([300, 3])
    add_points[:100, 0] = np.arange(100) / 10
    add_points[100:200, 1] = np.arange(100) / 10
    add_points[200:, 2] = np.arange(100) / 10
    points = np.concatenate([points, add_points])
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(points)
    P(save_path).parent.mkdir(exist_ok=True, parents=True)
    o3d.io.write_point_cloud(save_path, pcd)


def draw_points_with_color(points, colors, save_path='output.ply'):
    # need points
    points = points[:, :3]
    add_points = np.zeros([300, 3])
    add_points[:100, 0] = np.arange(100) / 10
    add_points[100:200, 1] = np.arange(100) / 10
    add_points[200:, 2] = np.arange(100) / 10
    points = np.concatenate([points, add_points])
    colors = np.concatenate([colors, np.zeros_like(add_points)])
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(points)
    pcd.colors = o3d.utility.Vector3dVector(colors)
    P(save_path).parent.mkdir(exist_ok=True, parents=True)
    o3d.io.write_point_cloud(save_path, pcd)


def load_pcd(pcd_path):
    lidar_data = o3d.t.io.read_point_cloud(pcd_path)
    lidar_points_positions = lidar_data.point.positions.numpy()
    lidar_points_intensity = lidar_data.point.intensity.numpy()
    lidar_points = np.concatenate([lidar_points_positions, lidar_points_intensity, np.zeros_like(lidar_points_intensity)], axis=1)
    return lidar_points


def write_pcd(points, save_path, intensity=False, color=False):
    pcd = o3d.t.geometry.PointCloud()
    device = o3d.core.Device("CPU:0")
    a = points.astype('float')
    dtype = o3d.core.float32
    pcd.point.positions = o3d.core.Tensor(a[:, :3], dtype)
    if intensity:
        pcd.point.intensity = o3d.core.Tensor(a[:, 3:4], dtype)
    if color:
        pcd.point.color = o3d.core.Tensor(a[:, 4:6], dtype)
    o3d.t.io.write_point_cloud(str(save_path), pcd)


def draw_occ(voxels):
    import torch
    def from_voxel_to_edges(voxels):
        bases_ = torch.arange(0, voxels.shape[0] * 8, 8)
        edges = torch.tensor([[0, 1], [1, 3], [3, 2], [2, 0], [4, 5], [5, 7], [7, 6], [6, 4], [0, 4], [1, 5], [2, 6], [3, 7]])  # lines along y-axis
        edges = edges.reshape((1, 12, 2)).repeat(voxels.shape[0], 1, 1)
        edges = np.array(edges + bases_[:, None, None])

        line_sets = o3d.geometry.LineSet()
        line_sets.points = o3d.open3d.utility.Vector3dVector(voxels.reshape((-1, 3)))
        line_sets.lines = o3d.open3d.utility.Vector2iVector(edges.reshape((-1, 2)))
        line_sets.paint_uniform_color((0, 0, 0))
        return voxels, line_sets
    voxel_grid, line_sets = from_voxel_to_edges(voxels)
    vis = o3d.visualization.Visualizer()
    opt = vis.get_render_option()
    opt.background_color = np.asarray([1, 1, 1])
    vis.add_geometry(voxel_grid)
    vis.add_geometry(line_sets)
    mesh_frame = o3d.geometry.TriangleMesh.create_coordinate_frame(size=0.5, origin=[0, 0, 2])
    vis.add_geometry(mesh_frame)
    o3d.visualization.draw_geometries([line_sets])


if __name__ == "__main__":
    draw_occ(np.array([[0,0,0], [3,3,3]]))




import numpy as np

def get_cameras(cali_file):
    with open(cali_file) as f:
        cali = yaml.safe_load(f)
    cameras = {}
    for cid in [cam]:
        cam_params = cali['rig'][cid]
        cameras[cid] = OpenCVFisheyeCamera(cam_params)
    return cameras








from mtv4d.utils.main import add

def test_add_integers():
    """ 测试 add 函数能否正确地将两个整数相加 """
    assert add(1, 2) == 3

def test_add_floats():
    """ 测试 add 函数能否正确地将两个浮点数相加 """
    assert add(1.5, 2.7) == 4.2




# 最后一次重复造轮子
from scipy.spatial.transform import Rotation
import os.path as op
import numpy as np
from pyquaternion import Quaternion
from pathlib import Path as P


def homo43(pts):
    return np.concatenate([pts, np.ones([len(pts), 1])], axis=1)


def to_corners_7(pts):
    """

    lidar cor:
       2 -------- 3
      /|         /|
    6 -------- 7 .
    | |        | |
    . 1 -------- 0
    |/         |/
    5 -------- 4

    """
    x, y, z = pts[:3]
    a, b, c = pts[3:6] / 2
    # u, v, w = pts[6:9]
    u, v, w = 0, 0, pts[6]
    corners = np.array(
        [
            [+a, -b, -c],
            [+a, +b, -c],
            [+a, +b, +c],
            [+a, -b, +c],
            [-a, -b, -c],
            [-a, +b, -c],
            [-a, +b, +c],
            [-a, -b, +c],
        ]
    )
    Rx = Rotation.from_euler("x", u).as_matrix()
    Ry = Rotation.from_euler("y", v).as_matrix()
    Rz = Rotation.from_euler("z", w).as_matrix()
    R = Rx @ Ry @ Rz
    corners = (R @ corners.T).T
    corners += np.array([x, y, z]).reshape(-1, 3)
    return corners


def to_corners_9(pts):
    """

    lidar cor:
       2 -------- 3
      /|         /|
    6 -------- 7 .
    | |        | |
    . 1 -------- 0
    |/         |/
    5 -------- 4

    """
    x, y, z = pts[:3]
    a, b, c = pts[3:6] / 2
    # u, v, w = pts[6:9]
    u, v, w = 0, 0, pts[6]
    corners = np.array(
        [
            [+a, -b, -c],
            [+a, +b, -c],
            [+a, +b, +c],
            [+a, -b, +c],
            [-a, -b, -c],
            [-a, +b, -c],
            [-a, +b, +c],
            [-a, -b, +c],
        ]
    )
    Rx = Rotation.from_euler("x", u).as_matrix()
    Ry = Rotation.from_euler("y", v).as_matrix()
    Rz = Rotation.from_euler("z", w).as_matrix()
    R = Rx @ Ry @ Rz
    corners = (R @ corners.T).T
    corners += np.array([x, y, z]).reshape(-1, 3)
    return corners


def draw_boxes(img, corners2d, text_list=None):
    if text_list is None:
        text_list = np.arange(len(corners2d.reshape(-1, 8, 2)))
    for idx, (pts, txt) in enumerate(zip(corners2d.reshape(-1, 8, 2),text_list)):
        pt = pts[:, :2].astype('int')
        # if (pt < 0).any() or (pt > 1600).any():  continue
        cv2.polylines(img, [pt[:4]], 2, (0, 255, 255), 2)
        cv2.polylines(img, [pt[4:]], 2, (0, 0, 255), 2)
        for i in range(4):
            cv2.line(img, tuple(pt[i]), tuple(pt[i + 4]), (0, 0, 255), 2)
        # if int(txt) < 20:
        if txt:
            cv2.putText(img, str(txt), tuple(pt[0]), 1, 2, (255, 255,0),2)
    return img


import cv2
import matplotlib.pyplot as plt

EPS_FLOAT32 = float(np.finfo(np.float32).eps)
MAX_FLOAT32 = float(np.finfo(np.float32).max)


def project_points_fisheye(points, T_c, inv_poly, image_size, intrinsic_mat, fov, focal):
    # 涉及到的参数，1 fov 2 inv_poly 3 intrinsic: focal /pp
    pts4d = homo43(points) @ T_c.T
    points = pts4d
    xc = points[:, 0]
    yc = points[:, 1]
    zc = points[:, 2]
    norm = np.sqrt(xc ** 2 + yc ** 2)
    theta = np.arctan2(norm, zc)
    fov_mask = theta > fov / 2 * np.pi / 180
    rho = (
            theta
            + inv_poly[0] * theta ** 3
            + inv_poly[1] * theta ** 5
            + inv_poly[2] * theta ** 7
            + inv_poly[3] * theta ** 9
    )
    width, height = image_size
    image_radius = np.sqrt((width / 2) ** 2 + (height) ** 2)
    rho[fov_mask] = 2 * image_radius / focal[0]
    xn = rho * xc / norm
    yn = rho * yc / norm
    xn[norm < EPS_FLOAT32] = 0
    yn[norm < EPS_FLOAT32] = 0
    norm_coords = np.stack([xn, yn, np.ones_like(xn)], axis=1)

    image_coords = norm_coords @ intrinsic_mat[:, :3].T
    return image_coords[:, :2]


def project_points_ordinary_cam(pts3d, T_c, intrinsic_mat):
    pts4d = homo43(pts3d) @ T_c.T
    corners3d = pts4d[:, :3] / pts4d[:, 2:3]
    pts_cam = (intrinsic_mat @ corners3d.T).T
    return pts_cam[:, :2]


def draw_points_to_cams(im, boxes, Tce, intrinsic_mat, inv_poly, save_im=True):
    """
    T_c   : T_cl  #
    boxes: 7个数
    """
    image_size = im.shape[:2]
    pts3d = np.concatenate([to_corners_7(b) for b in boxes])  # n8 * 3
    if inv_poly is None:
        pts2d = project_points_ordinary_cam(pts3d, Tce, intrinsic_mat)
    else:
        pts2d = project_points_fisheye(pts3d, Tce, inv_poly, im.shape[:2], intrinsic_mat, 200,
                                       [intrinsic_mat[0, 0], intrinsic_mat[1, 1]])
    draw_boxes(im, pts2d)
    from time import time
    path = f'/tmp/1234/3/{time()}.jpg'
    if save_im:
        result = cv2.imwrite(path, im)
    return im







from scipy.spatial import distance


def chamfer_distance(pts1, pts2):
    # points shape is num x 3
    dist_mat = distance.cdist(pts1, pts2, 'euclidean')
    valid_ab = dist_mat.min(-1).mean()
    valid_ba = dist_mat.min(-2).mean()

    return (valid_ba + valid_ab) / 2


import copy

import numpy as np
import scipy
from scipy.spatial.transform import Rotation as R, Rotation
from scipy.spatial import Delaunay
from pathlib import Path as P
import os.path as op



def in_hull(p, hull):
    """
    :param p: (N, K) test points
    :param hull: (M, K) M corners of a box
    :return (N) bool
    """
    try:
        if not isinstance(hull, Delaunay):
            hull = Delaunay(hull)
        flag = hull.find_simplex(p) >= 0
    except scipy.spatial.qhull.QhullError:
        print('Warning: not a hull %s' % str(hull))
        flag = np.zeros(p.shape[0], dtype=np.bool)
    return flag

def rotate_points_along_z(points, angle):
    """
    Args:
        points: (B, N, 3 + C)
        angle: (B), angle along z-axis, angle increases x ==> y
    Returns:

    """
    cosa = np.cos(angle)
    sina = np.sin(angle)
    zeros = np.zeros(points.shape[0])
    ones = np.ones(points.shape[0])
    rot_matrix = np.vstack([
        cosa, sina, zeros,
        -sina, cosa, zeros,
        zeros, zeros, ones
    ]).T.reshape(-1, 3, 3).astype(np.float32)
    points_rot = np.matmul(points[:, :, 0:3], rot_matrix)
    points_rot = np.concatenate([points_rot, points[:, :, 3:]], axis=-1)
    return points_rot


def boxes_to_corners_3d(boxes3d):
    """
    lidar cor:
           4 -------- 5
        /|         /|
        7 -------- 6 .
        | |        | |
        . 0 -------- 1
        |/         |/
        3 -------- 2
    Args:
        boxes3d:  (N, 7) [x, y, z, l, w, h, heading], (x, y, z) is the box center
    Returns:
        corners3d: (N, 8, 3)
    """
    template = np.array([
        [1, 1, -1],
        [1, -1, -1],
        [-1, -1, -1],
        [-1, 1, -1],
        [1, 1, 1],
        [1, -1, 1],
        [-1, -1, 1],
        [-1, 1, 1]
    ]) / 2
    corners3d = np.repeat(boxes3d[:, None, 3:6], 8, axis=1) * template
    corners3d = rotate_points_along_z(
        corners3d.reshape(-1, 8, 3), boxes3d[:, 6]).reshape(-1, 8, 3)
    corners3d += boxes3d[:, None, 0:3]

    return corners3d


def translate_yprt_from_world_to_frame_deprecated(box, Twe):
    # written by me and find the euler not that way
    x = box["psr"]["position"]["x"]
    y = box["psr"]["position"]["y"]
    z = box["psr"]["position"]["z"]
    x_ = box["psr"]["rotation"]["x"]
    y_ = box["psr"]["rotation"]["y"]
    z_ = box["psr"]["rotation"]["z"]
    R = Rotation.from_euler('zyx', [z_, y_, x_]).as_matrix()
    t = [x, y, z]
    Tow = np.eye(4)
    Tow[:3, :3] = R
    Tow[:3, 3] = t
    Toe = Tow @ Twe
    rotation_ego = Rotation.from_matrix(Tow[:3, :3]).as_euler('zyx')
    return {
        'position': Toe[:3, 3].tolist(),
        'scale': [box["psr"]["scale"]['x'], box["psr"]["scale"]['y'], box["psr"]["scale"]['z']],
        'rotation': rotation_ego.tolist()
    }


def translate_and_duplicate_pts3d_with_T(polyline, T):
    new_polyline = copy.deepcopy(polyline)
    new_polyline['vertices'] = translate_pts3d_with_T(new_polyline['vertices'], T)
    return new_polyline

def translate_pts3d_with_T(pts3d, T):
    pts3d = np.array(pts3d).reshape(-1, 3)
    if len(pts3d)==0:
        return pts3d
    pts4d = np.concatenate([pts3d, np.ones([pts3d.shape[0], 1])], axis=1)
    new_pts = (T@pts4d.T).T
    return new_pts[:, :3]

def translate_box9d_with_T(bx, Tts):
    # Tts 表示 从box的当前系转到目标系
    position = np.array([[bx["psr"]["position"]["x"], bx["psr"]
                        ["position"]["y"], bx["psr"]["position"]["z"], 1.0]])
    rotation = np.eye(4)
    rotation[:3, :3] = R.from_euler("XYZ", [bx["psr"]["rotation"]["x"], bx["psr"]["rotation"]["y"],
                                            bx["psr"]["rotation"]["z"]]).as_matrix()
    transformed_position = (position @ Tts.T).squeeze()
    transformed_rotation = R.from_matrix(
        (Tts @ rotation)[:3, :3]).as_euler("XYZ")
    transformed_box = copy.deepcopy(bx)
    transformed_box["psr"]["position"] = {
        "x": transformed_position[0],
        "y": transformed_position[1],
        "z": transformed_position[2],
    }
    transformed_box["psr"]["rotation"] = {
        "x": transformed_rotation[0],
        "y": transformed_rotation[1],
        "z": transformed_rotation[2],
    }
    return transformed_box

def translate_psr_with_T(psr, Tts):
    # Tts 表示 从box的当前系转到目标系
    position = np.array([[psr["position"]["x"],psr["position"]["y"], psr["position"]["z"], 1.0]])
    rotation = np.eye(4)
    rotation[:3, :3] = R.from_euler("XYZ", [psr["rotation"]["x"], psr["rotation"]["y"],
                                            psr["rotation"]["z"]]).as_matrix()
    transformed_position = (position @ Tts.T).squeeze()
    transformed_rotation = R.from_matrix(
        (Tts @ rotation)[:3, :3]).as_euler("XYZ")
    new_psr = copy.deepcopy(psr)
    new_psr["position"] = {
        "x": transformed_position[0],
        "y": transformed_position[1],
        "z": transformed_position[2],
    }
    new_psr["rotation"] = {
        "x": transformed_rotation[0],
        "y": transformed_rotation[1],
        "z": transformed_rotation[2],
    }
    return new_psr



def translate_psr_to_output_geometry(psr):
    output = {}
    output['pos_xyz'] = [psr['position']['x'], psr['position']['y'], psr['position']['z']]
    output['scale_xyz'] = [psr['scale']['x'], psr['scale']['y'], psr['scale']['z']]
    output['rot_xyz'] = [psr['rotation']['x'], psr['rotation']['y'], psr['rotation']['z']]
    return output
def translate_output_geometry_to_psr(box):
    output = {
        'position': {'x': box['pos_xyz'][0],
                     'y':box['pos_xyz'][1],
                     'z':box['pos_xyz'][2],
                     },
        'scale': {'x': box['scale_xyz'][0],
                     'y':box['scale_xyz'][1],
                     'z':box['scale_xyz'][2],
                     },
        'rotation': {'x': box['rot_xyz'][0],
                     'y':box['rot_xyz'][1],
                     'z':box['rot_xyz'][2],
                     },
    }
    return output

def read_pose_txt(pose_txt_path):
    with open(pose_txt_path, 'r') as f:
        data = f.readlines()
    time = []
    x = []
    y = []
    z = []
    q = []
    T = []
    frame = []
    for i, line in enumerate(data):
        d_line = line.split(' ')
        if d_line[0] == 'time':
            continue
        if len(time) > 0:
            if float(d_line[0]) == time[-1]:
                continue
        time.append(float(d_line[0]))
        x = float(d_line[1])
        y = float(d_line[2])
        z = float(d_line[3])
        q = np.array([float(d_line[4]), float(d_line[5]),
                     float(d_line[6]), float(d_line[7])])
        # frame.append(float(d_line[8]))
        frame.append(i)
        rot = R.from_quat(np.array(q)).as_matrix()
        T_ = np.eye(4)
        T_[:3, :3] = rot
        T_[:3, 3] = np.array([x, y, z])
        T.append(T_)
    Twes = np.stack(T, axis=0)
    return np.array(time), Twes, frame


def read_ego_paths(trajectory_path):
    def to_rotation(data):
        t = data[:3]
        R = Rotation.from_quat(data[3:7]).as_matrix()
        T = np.eye(4)
        T[:3, :3] = R
        T[:3, 3] = t
        return T

    a = np.loadtxt(trajectory_path)
    out_dict = {i[0]: to_rotation(i[1:8]) for i in a}
    timestamps = np.array(sorted(out_dict.keys()))
    return out_dict, timestamps






import sys
from xml.dom.expatbuilder import theDOMImplementation
sys.path.append(".")
from collections import defaultdict
from mtv4d.utils.io_base import read_json
import numpy as np
import os.path as op
from mtv4d.utils.misc_base import find_path_from_ts_and_dir_path_maybe_not_find
from pathlib import Path as P

# 这个脚本不被这个文件夹中其他内容引用！
def load_and_transfer_dn_boxes_from_frame(manual_frames_anno_dir, timestamps):
    dn_boxes_dict_ts2id, dn_boxes_dict_id2ts = defaultdict(dict), defaultdict(dict)
    if len(sorted(P(manual_frames_anno_dir).glob('*'))) == 0:
        return dn_boxes_dict_ts2id, dn_boxes_dict_id2ts
    for ts in timestamps:
        p = find_path_from_ts_and_dir_path_maybe_not_find(ts, manual_frames_anno_dir)
        # 主要时间花在这个find上面，可以考虑从timestamp直接获取文件名字。
        if p is not None:
            data = read_json(str(p))
            for box in data:
                box["ts"] = ts
                dn_boxes_dict_ts2id[ts][box["obj_track_id"]] = box
                dn_boxes_dict_id2ts[box["obj_track_id"]][ts] = box
    return dn_boxes_dict_ts2id, dn_boxes_dict_id2ts


def load_and_transfer_annos_from_map(manual_map_anno_dir):
    box_static_dict_id2info = {}
    map_boxes = read_json(manual_map_anno_dir)
    for i in map_boxes:
        box_static_dict_id2info[i["obj_track_id"]] = i
    return box_static_dict_id2info


def filter_ground_pts(scene_root, acc_pcl, ts, thres_dist=0.15):
    """
    acc_pcl是lidar系, ground_plane_model也是lidar系
    """
    file_name = find_path_from_ts_and_dir_path_maybe_not_find(ts, op.join(scene_root, "ground_plane_model"))
    if file_name is None:
        return acc_pcl[acc_pcl[:, 2] > -1.75, :3]
    abcd = np.loadtxt(file_name).reshape(4, 1)
    abcd = abcd * (2 * (abcd[2, 0] >= 0) - 1)
    acc_pcl_4d = np.concatenate([acc_pcl, np.ones([len(acc_pcl), 1])], axis=1)
    dist = acc_pcl_4d @ abcd
    acc_pcl = acc_pcl[dist.reshape(-1) > thres_dist]
    return acc_pcl


def find_near_points_index_of_visible_points(points, visible_inds, thres_near_point=0.1):
    from scipy.spatial.distance import cdist

    distances = cdist(points, points, "euclidean")
    distance_filtered = distances[visible_inds].reshape(len(visible_inds), len(points))
    idx = np.argsort(distance_filtered, axis=1)
    dist = np.take_along_axis(distance_filtered, idx, axis=1)
    return np.array(sorted(set(idx[dist < thres_near_point]))).astype("long")

def index_has_near_point_vis(points, visible_points, thres_near_point=0.1):
    from scipy.spatial.distance import cdist

    distances = cdist(points, visible_points, "euclidean")
    return np.arange(len(points))[distances.min(axis=1) < thres_near_point]


import numpy as np


# another attr is error list of tp samples

def cal_ap_list_from_tp_list(tp_list, fp_list, conf_list, num_ap_interval=101, is_cal_ap=False, draw_pr_curve=False):
    # tp_list means whether each sample is true in positive pred samples.
    npos = len(tp_list)
    tp = np.cumsum(tp_list).astype(float)
    fp = np.cumsum(fp_list).astype(float)
    conf = np.array(conf_list)
    prec = tp / (fp + tp)
    rec = tp / float(npos)
    rec_interp = np.linspace(0, 1, num_ap_interval)  # 101 steps, from 0% to 100% recall.
    precision = np.interp(rec_interp, rec, prec, right=0)
    confidence = np.interp(rec_interp, rec, conf, right=0)
    recall = rec_interp
    if is_cal_ap:
        ap = calc_ap(precision, 0.1, 0.1)
        return ap
    if draw_pr_curve:
        import matplotlib.pyplot as plt
        plt.plot(recall, precision), plt.plot(rec_interp, conf), plt.show()
    return precision, recall, confidence


def calc_ap(precision, min_recall=0.1, min_precision=0.1) -> float:
    """ Calculated average precision in nus """
    assert 0 <= min_precision < 1
    assert 0 <= min_recall <= 1
    prec = np.copy(precision)
    prec = prec[round(100 * min_recall) + 1:]  # Clip low recalls. +1 to exclude the min recall bin.
    prec -= min_precision  # Clip low precision
    prec[prec < 0] = 0
    return float(np.mean(prec)) / (1.0 - min_precision)



from setuptools import setup, find_packages


setup(
    name='mtv4d',
    version='0.1.9',
    packages=find_packages(exclude=['*scripts*']),
    description='A 4d data sdk',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='ioou',
    author_email='caacabr@163.com',
    url='https://github.com/ioou/mtv4d',
    install_requires=[
        "numpy",
        "scipy",
        "tqdm",
        "matplotlib",
        "pytest",
        "pyyaml",
    ],
)



import json
import pickle
import numpy as np
from pathlib import Path as P


def read_txt(path, to_format=True):
    with open(str(path)) as f:
        data = f.readlines()
    if to_format:
        data = [i.strip('\n') for i in data]
    return data


def write_txt(data, path, to_format=True):
    P(path).parent.mkdir(exist_ok=True, parents=True)
    if to_format:
        data = [i + '\n' for i in data]
    with open(str(path), 'w') as f:
        f.writelines(data)


def read_pickle(path):
    with open(str(path), 'rb') as f:
        data = pickle.load(f)
    return data


def write_pickle(data, path):
    P(path).parent.mkdir(exist_ok=True, parents=True)
    with open(str(path), 'wb') as f:
        pickle.dump(data, f)


def read_json(path):
    with open(str(path)) as f:
        data = json.load(f)
    return data


def format_floats(obj):
    if isinstance(obj, float):
        return round(obj, 4)
    elif isinstance(obj, dict):
        for k, v in obj.items():
            obj[k] = format_floats(v)
        return obj
    elif isinstance(obj, list):
        for i, v in enumerate(obj):
            obj[i] = format_floats(v)
        return obj
    else:
        return obj


def write_json(output_dict, save_path, format_float=False, indent=None):
    save_path = str(save_path)
    P(save_path).parent.mkdir(exist_ok=True, parents=True)
    if format_float:
        output_dict = format_floats(output_dict)
    with open(save_path, "w") as f:
        json.dump(output_dict, f, default=lambda x: round(x, 4), indent=indent)


def write_json_from_list(output_dict, save_path, format_float=False, indent=None, round_float=4):
    save_path = str(save_path)
    P(save_path).parent.mkdir(exist_ok=True, parents=True)
    if format_float:
        output_dict = format_floats(output_dict)
    with open(save_path, "w") as f:
        json.dump(output_dict, f, default=lambda x: round(x, round_float), indent=indent)


def read_points_from_pcd(lidar_path):
    import open3d as o3d
    lidar_data = o3d.t.io.read_point_cloud(lidar_path)
    lidar_points_positions = lidar_data.point.positions.numpy()
    lidar_points_intensity = lidar_data.point.intensity.numpy()
    lidar_points = np.concatenate([lidar_points_positions, lidar_points_intensity], axis=1)
    return lidar_points


def read_pypcd(path, intensity) -> np.ndarray:
    """read pcd file

    Parameters
    ----------
    path : Union[str, Path]
        _description_
    intensity : bool, optional
        _description_, by default True
        
    Returns
    -------
    np.ndarray
        - if intensity is false, return (N, 3) array, i.e. [[x,y,z], ...]
        - if intensity is true, return (N, 4) array, i.e. [[x,y,z,intensity], ...]
    """
    from pypcd_imp import pypcd
    pcd = pypcd.PointCloud.from_path(path)
    npdata = np.stack([pcd.pc_data['x'], pcd.pc_data['y'], pcd.pc_data['z']], axis=1)
    if intensity:
        npdata = np.concatenate([npdata, pcd.pc_data['intensity'].reshape(-1, 1)], axis=1)
    return npdata

def write_pypcd(npdata, pcd_save_path):
    from pypcd_imp import pypcd
    pc_data = np.array(
        [(0, 0, 0, 18), (0.5, -1.5, 2.5, 50)], dtype=[('x', '<f4'), ('y', '<f4'), ('z', '<f4'), ('intensity', '<f4')]
    )
    metadata = {
        'version': 0.7,
        'fields': ['x', 'y', 'z', 'intensity'],
        'size': [4, 4, 4, 4],
        'type': ['F', 'F', 'F', 'F'],
        'count': [1, 1, 1, 1],
        'width': 2,
        'height': 1,
        'viewpoint': [0, 0, 0, 1],
        'points': 2,
        'data': 'binary',
    }
    pcd = pypcd.PointCloud(metadata, pc_data)
    pcd.save_pcd(str(pcd_save_path), compression='binary')

def read_points_from_bin(path):
    return np.fromfile(str(path), dtype='float32').reshape(-1, 4)


def write_bin(data, path):
    P(path).parent.mkdir(exist_ok=True, parents=True)
    data.tofile(str(path))


def write_ply(pts, path, clrs=None):
    import open3d as o3d
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(pts)
    if clrs is not None:
        pcd.colors = o3d.utility.Vector3dVector(clrs)
    o3d.io.write_point_cloud(path, pcd)


def draw_points_with_coord_origin(points):
    import open3d as o3d
    add_points = np.zeros([300, 3])
    add_points[:100, 0] = np.arange(100) / 10
    add_points[100:200, 1] = np.arange(100) / 10
    add_points[200:, 2] = np.arange(100) / 10
    points = np.concatenate([points, add_points])
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(points)
    o3d.io.write_point_cloud("output.ply", pcd)



from pathlib import Path as P
import os.path as op
from mtv4d.annos_4d.helper import get_sync_filename, get_times, read_json_to_list, write_json_from_list
import numpy as np

cameras = [
    "camera1",
    "camera5",
    "camera8",
    "camera11",
    "camera2",
    "camera3",
    "camera4",
    "camera7",
    "camera6",
    "camera15",
    "camera12",
]


class Timestamps:
    def __init__(self, scene_root, cameras, data=None):
        self.scene_root = scene_root
        self.cameras = cameras
        if data is None:
            self.generate_time_json_with_none(scene_root, cameras)
        else:
            self.timestamp_list = data

    @classmethod
    def from_json(cls, scene_root, cameras):
        path = op.join(scene_root, "4d_anno_infos/ts_full.json")
        data = read_json_to_list(path)
        return cls(scene_root, cameras, data)

    def generate_time_json_with_none(self, scene_root, cameras):
        timestamps_traj = np.loadtxt(P(scene_root) / "whole_scene/ego-trajectory/trajectory_temp_horizontal.txt")
        lidar_path = P(scene_root) / "lidar/undistort_static_lidar1"
        timestamps, _ = get_times(lidar_path)
        output_ts_list = [{"lidar": ts} for ts in timestamps if ts in timestamps_traj]

        # overlapped_lidar1
        new_lidar1 = P(scene_root) / "lidar/overlapped_lidar1"
        new_lidar1 = new_lidar1 if new_lidar1.exists() else P(scene_root) / f"hidden_point_removal/occlusion_filter_lidar_lidar1"
        label_times, label_names = get_times(new_lidar1)
        [i.update({'overlapped_lidar1': get_sync_filename(label_names, label_times, i["lidar"])}) for i in output_ts_list]


        for cam in cameras:
            camera_path = P(scene_root) / "camera" / cam
            label_times, label_names = get_times(camera_path)
            [i.update({cam: get_sync_filename(label_names, label_times, i["lidar"])}) for i in output_ts_list]

            # camera_hpr_path = P(scene_root) / f"hidden_point_removal/occlusion_filter_lidar_{cam}"
            # label_times, label_names = get_times(camera_hpr_path)
            # [i.update({f"hpr_{cam}": get_sync_filename(label_names, label_times, i["lidar"])}) for i in output_ts_list]
        write_json_from_list(output_ts_list, op.join(scene_root, "4d_anno_infos/ts_full.json"))
        self.timestamp_list = output_ts_list
        

    @property
    def timestamps_dict(self):
        return {i["lidar"]: i for i in self.timestamp_list}

    @property
    def timestamps_dict_clean(self):
        return {i["lidar"]: i for i in self.timestamp_list if all(i.values())}

    @property
    def timestamps_given_cameras(self, cameras):
        return {i["lidar"]: [i for i in cameras] for i in self.timestamp_list if all(i.values())}

    @property
    def timestamps_full(self):
        return [i["lidar"] for i in self.timestamp_list]

    @property
    def timestamps_clean(self):
        print( len([i["lidar"] for i in self.timestamp_list if all(i.values())]))
        return [i["lidar"] for i in self.timestamp_list if all(i.values())]


if __name__ == "__main__":
    from time import time

    a = time()
    ts = Timestamps("/ssd1/data/4d/20230823_110018", cameras)
    print(time() - a)
    print(len(ts.timestamps_dict), len(ts.timestamps_dict_clean))
    a = time()
    ts = Timestamps.from_json("/ssd1/data/4d/20230823_110018", cameras)
    print(len(ts.timestamps_dict), len(ts.timestamps_dict_clean))
    print(time() - a)
